#include <cstdlib>
#include <iostream>
#include <string>
#include "asfheader.h"

using namespace std;

typedef struct guid
{
    unsigned int  v1;
    unsigned short v2;
    unsigned short v3;
    unsigned char v4[8];
} guid;

typedef struct asfguid
{
	guid ASFHeader;
	guid FileHeader;
	guid StreamHeader;
	guid AudioStream;
	guid VideoStream;
	guid CommentHeader;
	guid CodecCommentHeader;
	guid DataHeader;
	guid Index;
	guid Head1;
	guid Head2;
	guid NoErrorCorrection;
	guid AudioSpread;
	guid ExtendedComment;
	guid MarkerObject;
} asfguid;

struct ASFFILEHEADER {
	guid file_id; // stream GUID
	long long file_size;
	long long creation_time; //File creation time FILETIME 8
	long long num_packets;    //Number of packets UINT64 8
	long long play_duration; //Timestamp of the end position UINT64 8
	long long send_duration;  //Duration of the playback UINT64 8
	long long preroll; //Time to bufferize before playing UINT32 4
	int flags; //Unknown, maybe flags ( usually contains 2 ) UINT32 4
	unsigned int min_packet_size; //Min size of the packet, in bytes UINT32 4
	unsigned int max_packet_size; //Max size of the packet  UINT32 4
	unsigned int max_bitrate; //Maximum bitrate of the media (sum of all the stream)
};

static const asfguid ASFGUID =
{
    { 0x75B22630, 0x668E, 0x11CF, { 0xA6, 0xD9, 0x00, 0xAA, 0x00, 0x62, 0xCE, 0x6C }},
    { 0x8CABDCA1, 0xA947, 0x11CF, { 0x8E, 0xE4, 0x00, 0xC0, 0x0C, 0x20, 0x53, 0x65 }},
    { 0xB7DC0791, 0xA9B7, 0x11CF, { 0x8E, 0xE6, 0x00, 0xC0, 0x0C, 0x20, 0x53, 0x65 }},
    { 0xF8699E40, 0x5B4D, 0x11CF, { 0xA8, 0xFD, 0x00, 0x80, 0x5F, 0x5C, 0x44, 0x2B }},
    { 0xBC19EFC0, 0x5B4D, 0x11CF, { 0xA8, 0xFD, 0x00, 0x80, 0x5F, 0x5C, 0x44, 0x2B }},
    { 0x75B22633, 0x668E, 0x11CF, { 0xA6, 0xD9, 0x00, 0xAA, 0x00, 0x62, 0xCE, 0x6C }},
	{ 0x86D15240, 0x311D, 0x11D0, { 0xA3, 0xA4, 0x00, 0xA0, 0xC9, 0x03, 0x48, 0xF6 }},
	{ 0x75B22636, 0x668E, 0x11CF, { 0xA6, 0xD9, 0x00, 0xAA, 0x00, 0x62, 0xCE, 0x6C }},
	{ 0x33000890, 0xE5B1, 0x11CF, { 0x89, 0xF4, 0x00, 0xA0, 0xC9, 0x03, 0x49, 0xCB }},
    { 0x5FBF03B5, 0xA92E, 0x11CF, { 0x8E, 0xE3, 0x00, 0xC0, 0x0C, 0x20, 0x53, 0x65 }},
    { 0xABD3D211, 0xA9BA, 0x11CF, { 0x8E, 0xE6, 0x00, 0xC0, 0x0C, 0x20, 0x53, 0x65 }},
	{ 0x20FB5700, 0x5B55, 0x11CF, { 0xA8, 0xFD, 0x00, 0x80, 0x5F, 0x5C, 0x44, 0x2B }},
	{ 0xBFC3CD50, 0x618F, 0x11CF, { 0x8B, 0xB2, 0x00, 0xAA, 0x00, 0xB4, 0xE2, 0x20 }},
	{ 0xD2D0A440, 0xE307, 0x11D2, { 0x97, 0xF0, 0x00, 0xA0, 0xC9, 0x5E, 0xA8, 0x50 }},
	{ 0xF487CD01, 0xA951, 0x11CF, { 0x8E, 0xE6, 0x00, 0xC0, 0x0C, 0x20, 0x53, 0x65 }}
};

struct BITMAPINFOHEADER *vidheader = NULL;
struct WAVEFORMAT *audheader = NULL;

char vidbuf[MAX_BUFSIZE];
char audbuf[MAX_BUFSIZE];
int vidbufsize = 0;
int audbufsize = 0;
int iskeyframe = 0;
int vidtime = 0;
int vidoffs = 0;
int audoffs = 0;

char *audextdata = NULL;
int audextsize = 0;
char *vidextdata = NULL;
int vidextsize = 0;

int stream_vid = 0;
int stream_aud = 0;

uint16_t title[1024];
uint16_t author[1024];
int playtime;

#define STREAM_BUFFER 8192
static unsigned char readbuf[STREAM_BUFFER];
static int readpos = 0;
static int readlen = 0;

static FILE *f_asf = NULL;

static void printGUIDType(guid x) {
	if(memcmp(&x, &ASFGUID.ASFHeader, sizeof(guid)) == 0) printf("ASFHeader");
	else if(memcmp(&x, &ASFGUID.FileHeader, sizeof(guid)) == 0) printf("FileHeader");
	else if(memcmp(&x, &ASFGUID.StreamHeader, sizeof(guid)) == 0) printf("StreamHeader");
	else if(memcmp(&x, &ASFGUID.AudioStream, sizeof(guid)) == 0) printf("AudioStream");
	else if(memcmp(&x, &ASFGUID.VideoStream, sizeof(guid)) == 0) printf("VideoStream");
	else if(memcmp(&x, &ASFGUID.CommentHeader, sizeof(guid)) == 0) printf("CommentHeader");
	else if(memcmp(&x, &ASFGUID.CodecCommentHeader, sizeof(guid)) == 0) printf("CodecCommentHeader");
	else if(memcmp(&x, &ASFGUID.DataHeader, sizeof(guid)) == 0) printf("DataHeader");
	else if(memcmp(&x, &ASFGUID.Index, sizeof(guid)) == 0) printf("Index");
	else if(memcmp(&x, &ASFGUID.Head1, sizeof(guid)) == 0) printf("Head1");
	else if(memcmp(&x, &ASFGUID.Head2, sizeof(guid)) == 0) printf("Head2");
	else if(memcmp(&x, &ASFGUID.NoErrorCorrection, sizeof(guid)) == 0) printf("NoErrorCorrection");
	else if(memcmp(&x, &ASFGUID.AudioSpread, sizeof(guid)) == 0) printf("AudioSpread");
	else if(memcmp(&x, &ASFGUID.ExtendedComment, sizeof(guid)) == 0) printf("ExtendedComment");
	else if(memcmp(&x, &ASFGUID.MarkerObject, sizeof(guid)) == 0) printf("MarkerObject");
	else printf("Unknown");
}

static void PrintFourCC(int FourCC)
{
	char chars[5];
	chars[0] = (char)(FourCC & 0xFF);
	chars[1] = (char)((FourCC >> 8) & 0xFF);
	chars[2] = (char)((FourCC >> 16) & 0xFF);
	chars[3] = (char)((FourCC >> 24) & 0xFF);
	chars[4] = 0;
	
	printf("%s", chars);
}

static std::string getFourCC(int FourCC)
{
	char chars[5];
	std::string s;

	chars[0] = (char)(FourCC & 0xFF);
	chars[1] = (char)((FourCC >> 8) & 0xFF);
	chars[2] = (char)((FourCC >> 16) & 0xFF);
	chars[3] = (char)((FourCC >> 24) & 0xFF);
	chars[4] = 0;
	
	s = chars;	
	return s;
}

#    ifndef mmioFOURCC
#        define mmioFOURCC( ch0, ch1, ch2, ch3 )                \
        ( (unsigned int)(char)(ch0) | ( (unsigned int)(char)(ch1) << 8 ) |  \
        ( (unsigned int)(char)(ch2) << 16 ) | ( (unsigned int)(char)(ch3) << 24 ) )
#    endif

static int getLowerFourCC(int FourCC)
{
	int i;
	char chars[4];
	
	for(i=0;i<4;i++) {
		chars[i] = (char)(FourCC & 0xFF);
		if((chars[i] >= 'A') && (chars[i] <= 'Z'))
			chars[i] += 'a' - 'A';
		FourCC >>= 8;
	}

	return mmioFOURCC(chars[0], chars[1], chars[2], chars[3]);
}

static std::string getAudioCodec(unsigned short wFormatTag)
{
	std::string s;
	switch(wFormatTag)
	{
		case 0x0000 : s = "Unknown"; break;
		case 0x0001 : s = "PCM"; break;
		case 0x0002 : s = "ADPCM"; break;

		case 0x0050 : s = "MPEG"; break;
		case 0x0055 : s = "MPEG 1 Layer 3(MP3)"; break;

		// http://msdn.microsoft.com/library/default.asp?url=/library/en-us/wmform/htm/mediatypeidentifiers.asp
		case 0x0160 : s = "Window Media Audio v1"; break;
		case 0x0161 : s = "Window Media Audio v2 (7/8/9)"; break;
		case 0x0162 : s = "Window Media Audio 9 Professional"; break;
		case 0x0163 : s = "Window Media Audio 9 Lossless"; break;
		case 0x000A : s = "Window Media Audio 9 Speech"; break;

		case 0x2000 : s = "AC3"; break;
		case 0x2001 : s = "DTS"; break;

		// Vorbis
		case 0x674f	: s = "Vorbis 1"; break;
		case 0x6750	: s = "Vorbis 2"; break;
		case 0x6751	: s = "Vorbis 3"; break;
		case 0x676f	: s = "Vorbis 1+"; break;
		case 0x6770	: s = "Vorbis 2+"; break;
		case 0x6771	: s = "Vorbis 3+"; break;

		default :
			char tmp[256];
			sprintf(tmp, "Unknown(code:0x%x)", wFormatTag); 
			s = tmp;
			break;
	}
	return s;
}

static std::string getVideoCodec(int fourcc)
{
	std::string s;
	int FourCC = getLowerFourCC(fourcc);
	
	switch(FourCC)
	{
		case mmioFOURCC('w', 'm', 'v', '1'): s = "Window Media Video v1/7"; break;
		case mmioFOURCC('w', 'm', 'v', '2'): s = "Window Media Video v2/8"; break;
		case mmioFOURCC('w', 'm', 'v', '3'): s = "Window Media Video v3/9"; break;
		
		default :
			s = "Unknown(" + getFourCC(FourCC) + ")";
			break;
	}
	return s;
}

#if 0
#define printGUID(x) { \
		int i; \
		printf("%08x, %04x, %04x", x.v1, x.v2, x.v3); \
		for(i=0;i<8;i++) \
			printf(", %02x", x.v4[i]); \
		printf(" ("); \
		printGUIDType(x); \
		printf(")"); \
		printf("\n"); \
	}
#else
#define printGUID(x) { \
		printGUIDType(x); \
		printf("\n"); \
	}
#endif

static inline int ReadData(int size, void *data)
{
	int len = readlen - readpos;
	if(len > 0) {
		if(len > size) len = size;
		memcpy(data, readbuf + readpos, len);
		data = (char *)data + len;
		size -= len;
		readpos += len;
	}
	while(size > 0) {
		len = fread(data, 1, size, f_asf);
		size -= len;
		if(len == 0) {
			printf("read error!!!\n");
			return 0;
		}
	}
	return 1;
}

static inline void SkipData(int size)
{
	int len = readlen - readpos;
	if(size <= len) {
		readpos += size;
	} else {
		fseek(f_asf, size - len, SEEK_CUR);
		readlen = readpos = 0;
	}
}

static inline int FillBuffer()
{
	int len = readlen - readpos;
	if(len > 0) {
		return len;
	} else {
		readlen = fread(readbuf, 1, STREAM_BUFFER, f_asf);
		readpos = 0;
		return readlen;
	}
}

static inline int GetPos()
{
	int ret = ftell(f_asf);
	ret -= readlen - readpos;
	return ret;
}

static inline int ReadByte()
{
	return ((readlen - readpos) > 0) ? readbuf[readpos++] :
		(FillBuffer() > 0 ? readbuf[readpos++] : -256);
}

static inline int ReadWord()
{
	int ret = ReadByte();
	return ret | (ReadByte() << 8);
}

static inline int ReadInt()
{
	int ret = ReadByte();
	ret |= ReadByte() << 8;
	ret |= ReadByte() << 16;
	ret |= ReadByte() << 24;
	return ret;
}

static long long datasize;

static inline int ReadValue(int lentype)
{
	switch(lentype & 0x03) {
		case 1: datasize--; return ReadByte();
		case 2: datasize-=2; return ReadWord();
		case 3: datasize-=4; return ReadInt();
	}
	return 0;
}

static int packetmax, packetmin;
static int payloads = 0;
static int payloadlentype = 0;
static int payloadlen = 0;					
static long long numpacket;
static unsigned char types[7];
static unsigned int lens[7] = {0, };
static int packetbegin;
static int preroll;

int asfOpen(const char *name)
{
	int fsize, fpos;
	long long size;

	if(f_asf) {
		printf("file already opened.\n");
		return -1;
	}

	f_asf = fopen(name, "rb");
	
	readpos = readlen = 0;

	if(vidheader) free(vidheader);
	vidheader = NULL;
	vidbufsize = 0;
	vidoffs = 0;

	if(audheader) free(audheader);
	audheader = NULL;
	audbufsize = 0;
	audoffs = 0;
	
	numpacket = 0;
	packetbegin = 0;
	
	title[0] = 0;
	author[0] = 0;

	if(f_asf == NULL) {
		printf("file open failed.\n");
		return -1;
	}

	fseek(f_asf, 0, SEEK_END);
	fsize = ftell(f_asf);	
	printf("File Size: %d\n", fsize);
	fseek(f_asf, 0, SEEK_SET);
	
	while(1) {
		guid GUID;

		if(!ReadData(16, &GUID)) break;
		if(!ReadData(8, &size)) break;

		printf("File GUID: ");printGUID(GUID);
		printf("Size: %d\n", size);
		size -= 24;

		if(memcmp(&GUID, &ASFGUID.ASFHeader, sizeof(guid)) == 0) {
			int objcount;
			guid objguid;
			long long objsize;

			objcount = ReadInt();
			SkipData(2);
			size -= 6;
			printf("ASF Header\n");
			printf("Number of object: %d\n", objcount);

			while(objcount) {
				ReadData(16, &objguid);
				ReadData(8, &objsize);
				size -= objsize;

				printf("Object GUID: ");printGUID(objguid);
				printf("Object Size: %d\n", objsize);
				objsize -= 24;

				if(memcmp(&objguid, &ASFGUID.FileHeader, sizeof(guid)) == 0) {
					struct ASFFILEHEADER fh;
					long long time;
	 				int h, m, s;

					ReadData(sizeof(fh), &fh);
					objsize -= sizeof(fh);

					printf("--------------------------------------------------\n");
					printf("File ID = ");printGUID(fh.file_id);
					printf("File Size = %ld\n", fh.file_size);
					printf("Packet Count = %ld\n", fh.num_packets);
					//printf("Play Duration = %ld\n", fh.play_duration);
					time = (fh.play_duration) / 100000L;
					h = time / 360000;
					time -= h * 360000;
					m = time / 6000;
					time -= m * 6000;
					s = time / 100;
					time -= s * 100;
					printf("Play Time = %02d:%02d:%02d.%02d\n", h, m, s, time);
					//printf("Send Duration = %ld\n", fh.send_duration);
					time = (fh.send_duration) / 100000L;
					playtime = time / 100;
					h = time / 360000;
					time -= h * 360000;
					m = time / 6000;
					time -= m * 6000;
					s = time / 100;
					time -= s * 100;
					printf("Send Time = %02d:%02d:%02d.%02d\n", h, m, s, time);
					printf("Preroll = %ld\n", fh.preroll);
					printf("Flags = %x\n", fh.flags);
					printf("Minimum Packet Size = %d\n", fh.min_packet_size);
					printf("Maximum Packet Size = %d\n", fh.max_packet_size);
					printf("Maximum Bitrate = %d\n", fh.max_bitrate);
					printf("--------------------------------------------------\n");
					packetmax = fh.max_packet_size;
					packetmin = fh.min_packet_size;
					preroll = fh.preroll;
				} else if(memcmp(&objguid, &ASFGUID.StreamHeader, sizeof(guid)) == 0) {
					guid streamguid;
					guid errorguid;
					long long timeOffset;
					int streamlen, errorlen;
					unsigned short stream_flags;
					
					ReadData(16, &streamguid);
					ReadData(16, &errorguid);
					ReadData(8, &timeOffset);
					ReadData(4, &streamlen);
					ReadData(4, &errorlen);
					ReadData(2, &stream_flags);
					SkipData(4);
					objsize -= 54;

					printf("Stream Type: ");printGUID(streamguid);
					printf("Error Type: ");printGUID(errorguid);
					printf("Time Offset: %ld\n", timeOffset);
					printf("Stream Length: %d\n", streamlen);
					printf("Error Length: %d\n", errorlen);
					printf("Flags: %04x\n", stream_flags);
					
					if(memcmp(&streamguid, &ASFGUID.AudioStream, sizeof(guid)) == 0) {
						if(audheader == NULL) {
							audheader = (struct WAVEFORMAT *)malloc(objsize);
							audextdata = (char *)audheader + sizeof(struct WAVEFORMAT);
							ReadData(objsize, audheader);
							audextsize = audextdata[0] | audextdata[1] << 8;
							audextdata += 2;
							objsize = 0;
	
							stream_aud = stream_flags;
							
							printf("--------------------------------------------------\n");
							printf("wFormatTag = %x\n", audheader->wFormatTag);
							std::string codec = getAudioCodec(audheader->wFormatTag);
							printf("Audio Codec = %s\n", codec.c_str());
							printf("nChannels = %d\n", audheader->nChannels);
							printf("nSamplesPerSec = %d\n", audheader->nSamplesPerSec);
							printf("nAvgBytesPerSec = %d\n", audheader->nAvgBytesPerSec);
							printf("nBlockAlign = %d\n", audheader->nBlockAlign);
							printf("wBitsPerSample = %d\n", audheader->wBitsPerSample);
							printf("--------------------------------------------------\n");
							printf("Audio Stream Bitrate = %d kbps\n", audheader->nAvgBytesPerSec * 8 / 1000);
							printf("Audio Extension Info Length = %d\n", audextsize);
							printf("--------------------------------------------------\n");
						} else {
							printf("--------------------------------------------------\n");
							printf("Skip Audio Stream\n");
							printf("--------------------------------------------------\n");
						}
					} else if(memcmp(&streamguid, &ASFGUID.VideoStream, sizeof(guid)) == 0) {
						if(vidheader == NULL) {
							unsigned int width, height;
							unsigned char flags;
							unsigned short bmsize;
	
							ReadData(4, &width);
							ReadData(4, &height);
							ReadData(1, &flags);
							ReadData(2, &bmsize);
							vidheader = (struct BITMAPINFOHEADER *)malloc(bmsize);
							vidextdata = (char *)vidheader + sizeof(struct BITMAPINFOHEADER);
							vidextsize = bmsize - sizeof(struct BITMAPINFOHEADER);
							ReadData(bmsize, vidheader);
							objsize -= 11 + bmsize;

							stream_vid = stream_flags;

							printf("--------------------------------------------------\n");
							printf("ASFVideo.Width = %d\n", width);
							printf("ASFVideo.Height = %d\n", height);
							printf("ASFVideo.ReservedFlags = %d\n", flags);
							printf("ASFVideo.HeaderSize = %d\n", bmsize);
							printf("--------------------------------------------------\n");
							printf("biSize = %d\n", vidheader->biSize);
							printf("biWidth = %d\n", vidheader->biWidth);
							printf("biHeight = %d\n", vidheader->biHeight);
							printf("biPlanes = %d\n", vidheader->biPlanes);
							printf("biBitCount = %d\n", vidheader->biBitCount);
							printf("biCompression = ");PrintFourCC(vidheader->biCompression);printf("\n");
							std::string codec = getVideoCodec(vidheader->biCompression);
							printf("Video Codec = %s\n", codec.c_str());
							printf("biSizeImage = %d\n", vidheader->biSizeImage);
							printf("biXPelsPerMeter = %d\n", vidheader->biXPelsPerMeter);
							printf("biYPelsPerMeter = %d\n", vidheader->biYPelsPerMeter);
							printf("biClrUsed = %d\n", vidheader->biClrUsed);
							printf("biClrImportant = %d\n", vidheader->biClrImportant);
							printf("--------------------------------------------------\n");
							printf("Video Extension Info Length = %d\n", vidextsize);
							printf("Skip Video Stream = %d\n", objsize);
							printf("--------------------------------------------------\n");
						} else {
							printf("--------------------------------------------------\n");
							printf("Skip Video Stream\n");
							printf("--------------------------------------------------\n");
						}
					}
				} else if(memcmp(&objguid, &ASFGUID.CommentHeader, sizeof(guid)) == 0) {
					int titlelen, authorlen, copyrightlen, commentlen, urllen;
					titlelen = ReadWord();
					authorlen = ReadWord();
					copyrightlen = ReadWord();
					commentlen = ReadWord();
					urllen = ReadWord();
					objsize -= 10;
					ReadData(titlelen, title);objsize -= titlelen;
					title[titlelen] = 0;
					ReadData(authorlen, author);objsize -= authorlen;
					author[authorlen] = 0;
					// Skip else...			
				}

				SkipData(objsize);
				objcount--;
			}
		} else if(memcmp(&GUID, &ASFGUID.DataHeader, sizeof(guid)) == 0) {
			guid fileid;
			unsigned short reserved;

			ReadData(16, &fileid);
			ReadData(8, &numpacket);
			ReadData(2, &reserved);
			size -= 26;

			printf("File ID: ");printGUID(fileid);
			printf("Packet Count = %d\n", numpacket);
			printf("Reserved Flags = %04x\n", reserved);

			if( (size <= 0) || (numpacket <= 0) ) {
				printf("Uncounted Packet...\n");
				size = fsize - GetPos();
				numpacket = -1;
			}

			payloads = 0;
			payloadlen = 0;
			datasize = size;
			lens[1] = 0;

			break;
		} else {
			break;
		}

		printf("------------------------------------ Skip: %d\n", size);
		SkipData(size);
	}

    return numpacket;
}

static int paysize, paytime, streamno;
static unsigned char ptimedelta;
static int compressed;
static int compressedtime;

int asfRead()
{
	int filled = 0;

	if(!f_asf) return 0;
	if(numpacket == 0) return 0;

	while((datasize > 0) && numpacket && !filled) {
		if((payloads == 0) && (payloadlen == 0)) {
			// read Payloads
			static const char *LENGTHTYPE[4] = {"None", "Byte", "Word", "DWord"};
			unsigned char error, lentype, property;
			unsigned short duration;
			int sendtime;
			int i;
			
			if(packetbegin) {
				//printf("Padding: %d\n", lens[1]);
				SkipData(lens[1]);
				datasize -= lens[1];
				if(numpacket > 0) numpacket--;
				if((!numpacket) || (datasize <= 0)) {
					printf("Data Size Over...\n");
					numpacket = 0;
					return 0;
				}
			}

			//printf("==================================================\n");
			packetbegin = datasize;
			
			error = ReadByte();datasize--;

			if(error & 0x80) {
				if( (error != 0x82) && (error != 0x90) ) {
					printf("Error Correction Data Length: %d\n", error & 0x0F);
					printf("Opaque Data Present: %s\n", error & 0x10 ? "Yes" : "No");
					printf("Error Correction Length Type: %d\n", (error >> 5) & 0x03);
					printf("Error Correction Present: %s\n", error & 0x80 ? "Yes" : "No");
					printf("--------------------------------------------------\n");
					printf("Error Correction Object is Broken(%d)\n", numpacket);
					numpacket = 0;
					break;
				}
				SkipData(error & 0x0F);
				datasize -= error & 0x0F;
				lentype = ReadByte();datasize--;
			} else {
				lentype = error;
			}
			
			property = ReadByte();datasize--;
			
			types[0] = (lentype >> 1) & 0x03;
			types[1] = (lentype >> 3) & 0x03;
			types[2] = (lentype >> 5) & 0x03;
			
			types[3] = (property >> 0) & 0x03;
			types[4] = (property >> 2) & 0x03;
			types[5] = (property >> 4) & 0x03;
			types[6] = (property >> 6) & 0x03;
			
			/*
			printf("Multiple Payloads Present: %s\n", lentype & 1 ? "Yes" : "No");
			printf("Sequence Type: %s\n", LENGTHTYPE[types[0]]);
			printf("Padding Length Type: %s\n", LENGTHTYPE[types[1]]);
			printf("Packet Length Type: %s\n", LENGTHTYPE[types[2]]);
			printf("--------------------------------------------------\n");
			printf("Replicated Data Length Type: %s\n", LENGTHTYPE[types[3]]);
			printf("Offset Into Media Object Length Type: %s\n", LENGTHTYPE[types[4]]);
			printf("Media Object Number Length Type: %s\n", LENGTHTYPE[types[5]]);
			printf("Stream Number Length Type: %s\n", LENGTHTYPE[types[6]]);
			printf("Error Correction Present: %s\n", lentype & 0x80 ? "Yes" : "No");
			printf("--------------------------------------------------\n");
			*/

			if(types[6] != 1) {
				printf("Broken Packet...(%d)\n", numpacket);
				numpacket = 0;
				break;
			}

			for (i=0; i < 3; i++) {
				lens[i] = ReadValue(types[i]);
				/*
				switch(i) {
					case 0:					
						printf("Sequence: %d (%s)\n", lens[i], LENGTHTYPE[types[0]]);
						break;
					case 1:
						printf("Padding Length: %d (%s)\n", lens[i], LENGTHTYPE[types[1]]);
						break;
					case 2:
						printf("Packet Length: %d (%s)\n", lens[i], LENGTHTYPE[types[2]]);
						break;
				}
				/*/
			}

			if(packetmax == packetmin) {
				lens[2] = packetmax;
			} else if(types[2] == 0) {
				lens[2] = packetmax;
			}

			if(lens[2] < packetmin) {
				lens[1] += packetmin - lens[2];
			}

			//printf("Packet Length: %d\n", lens[2]);

			sendtime = ReadInt();
			duration = ReadWord();
			datasize -= 6;

			//printf("Send Time: %d\n", sendtime);
			//printf("Duration: %d\n", duration);
			//printf("FPS: %.3f\n", 1000.0 / duration);

			if(lentype & 1) {
				unsigned char multipay = ReadByte();datasize--;
				payloads = multipay & 0x3F;
				payloadlentype = multipay >> 6;
				//printf("Number of Payloads: %d\n", payloads);
				//printf("Payload Length Type: %s\n", LENGTHTYPE[payloadlentype]);
			} else {
				payloads = 1;
				payloadlentype = 0;
				//printf("Single Payload...\n");
			}
		} else if (payloads < 0) {
			printf("Error!!!\n");
			numpacket = 0;
			return 0;
		} else if (payloadlen == 0) {
			int i;

			for (i=6; i > 2; i--) {
				lens[i] = ReadValue(types[i]);
			}
			
			//printf("Stream Number: %d (%s)\n", lens[6] & 0x7F, (lens[6] & 0x7F) == stream_aud ? "Audio" : "Video");

			/*
			printf("-------------------------------------------------- %d\n", payloads);
			printf("Stream Number: %d\n", lens[6] & 0x7F);
			printf("Key Frame: %s\n", lens[6] & 0x80 ? "Yes" : "No");
			printf("Media Object Number: %d\n", lens[5]);
			if(lens[3] == 1) {
				printf("Presentation Time: %d\n", lens[4]);
			} else {
				printf("Offset Into Media Object: %d\n", lens[4]);
			}
			printf("Replicated Data Length: %d\n", lens[3]);
			*/
			
			streamno = lens[6] & 0x7F;
			compressed = (lens[3] == 1);

			if(compressed) {
				ptimedelta = ReadByte();datasize--;
				//printf("Compressed Payload\n");
				//printf("Presentation Time Delta: %d\n", ptimedelta);
				compressedtime = lens[4];
				//printf("Payload Time: %d\n", paytime);
				//numpacket = 0;
				//return 0;
			} else {
				if(lens[3] < 8) {
					printf("Replicated Data is Error!!!\n");
					numpacket = 0;
					return 0;
				} else {
					paysize = ReadInt();
					paytime = ReadInt();
					//printf("Payload Size: %d\n", paysize);
					//printf("Payload Time: %d\n", paytime);
					lens[3] -= 8;
					datasize -= 8;
				}
				SkipData(lens[3]);
				datasize -= lens[3];
			}

			if(payloadlentype > 0) {
				payloadlen = ReadValue(payloadlentype);
			} else {
				if(!compressed) lens[1] = 0; // Remove padding area.
				payloadlen = lens[2] - (packetbegin - datasize) - lens[1];
			}
			//printf("Payload Length: %d\n", payloadlen);
		} else {
			int readsize, payoffs;

			if(compressed) {
				readsize = ReadByte();datasize--;payloadlen--;
				if(readsize > payloadlen) {
					readsize = payloadlen;
					streamno = 0;
				}
				paysize = readsize;
				paytime = compressedtime;
				//printf("Packet Size: %d\n", paysize);
				//printf("Packet Time: %d\n", paytime);
				compressedtime += ptimedelta;
				payoffs = 0;
			} else {
				readsize = payloadlen;
				payoffs = lens[4];
			}
			
			//printf("Packet Length: %d\n", paysize);			
			//printf("Payload Offset: %d\n", payoffs);
			//printf("Payload Length: %d (%d)\n", readsize, payoffs + readsize);

			if(streamno == stream_vid) {
				if(vidoffs >= payoffs) {
					ReadData(readsize, &vidbuf[payoffs]);
					vidoffs = payoffs + readsize;
				} else {
					vidoffs = 0;
					printf("Packet Error(Ignore).\n");
					SkipData(readsize);
				}
				if(vidoffs >= paysize) {
					vidbufsize = paysize;
					iskeyframe = lens[6] & 0x80;
					vidtime = paytime - preroll;
					filled = 1;
				}
			} else if(streamno == stream_aud) {
				ReadData(readsize, &audbuf[payoffs]);
				if((readsize + payoffs) >= paysize) {
					audbufsize = paysize;
					filled = 1;
					//printf("Audio Packet Filled(%d).\n", paysize);
				}
			} else {
				SkipData(readsize);
			}

			payloadlen -= readsize;
			datasize -= readsize;

			if(payloadlen == 0) {
				payloads--;

				if((payloadlentype > 0) && (payloads == 0)) {
					//int pad = lens[1];
					lens[1] = lens[2] - (packetbegin - datasize);
					//if(pad != lens[1])
					//	printf("Padding: %d -> %d\n", pad, lens[1]);
					if(lens[1] < 0) {
						numpacket = 0;
						printf("Packet Length Error!!!\n");
						break;
					}
				}
			}
		}
	}
	
	if(datasize <= 0) numpacket = 0;	
	return numpacket;
}

void asfClose()
{
	if(f_asf) fclose(f_asf);
	f_asf = NULL;

	if(vidheader) free(vidheader);
	vidheader = NULL;
	vidbufsize = 0;

	if(audheader) free(audheader);
	audheader = NULL;
	audbufsize = 0;
	
	numpacket = 0;
	packetbegin = 0;
}

void asfVideoOff()
{
	stream_vid = 0;
}
