/**
 * @file common.h
 * common internal api header.
 */

#ifndef COMMON_H
#define COMMON_H

#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <math.h>
#include <stddef.h>
#include <float.h>

#ifdef DEBUG
#  define STIN static
#else
#  ifdef __GNUC__
#    define STIN static __inline__
#  elif _WIN32
#    define STIN static __inline
#  else
#    define STIN static
#  endif
#endif

#include <inttypes.h>

#ifndef INT64_MAX
#define INT64_MAX 9223372036854775807LL
#endif

#ifndef int64_t_C
#define int64_t_C(c)     (c ## LL)
#define uint64_t_C(c)    (c ## ULL)
#endif

#include "bswap.h"
//#include "bitstrm.h"

#define TUNECPU generic
#undef HAVE_MMX
#define __CPU__ 586
#define HAVE_BUILTIN_VECTOR 1
#define CONFIG_DECODERS 1
#define ALT_BITSTREAM_READER

#ifndef M_PI
#define M_PI    3.14159265358979323846
#endif

#ifndef ENODATA
#define ENODATA  61
#endif

#ifndef offsetof
#define offsetof(T,F) ((unsigned int)((char *)&((T *)0)->F))
#endif

#if defined(__GNUC__) && (__GNUC__ > 3 || __GNUC__ == 3 && __GNUC_MINOR__ > 0)
#    define always_inline __attribute__((always_inline)) inline
#else
#    define always_inline inline
#endif

#ifndef DEBUG
#define NDEBUG
#endif

/* dprintf macros */
#ifdef DEBUG
#  define dprintf printf
#else
#  define dprintf(fmt,...)
#endif

//#define av_abort()      do { printf("Abort at %s:%d\n", __FILE__, __LINE__); exit(0); } while (0)

#ifdef ARCH_X86
// avoid +32 for shift optimization (gcc should do that ...)
static inline uint32_t NEG_USR32(uint32_t a, int8_t s){
    asm ("shrl %1, %0\n\t"
         : "+r" (a)
         : "ic" ((uint8_t)(-s))
    );
    return a;
}
#else
#    define NEG_USR32(a,s) (((uint32_t)(a))>>(32-(s)))
#endif

/* bit input */

#define VLC_TYPE int16_t

typedef struct VLC {
    int bits;
    VLC_TYPE (*table)[2]; ///< code, bits
    int table_size, table_allocated;
} VLC;

typedef struct RL_VLC_ELEM {
    int16_t level;
    int8_t len;
    uint8_t run;
} RL_VLC_ELEM;

int init_vlc(VLC *vlc, int nb_bits, int nb_codes,
             const void *bits, int bits_wrap, int bits_size,
             const void *codes, int codes_wrap, int codes_size);
void free_vlc(VLC *vlc);

//#define TCPMP
#undef TCPMP

#ifdef TCPMP
#define GetBitContext bitstream
#define get_bits(p, n) bitget(p, n)
#define skip_bits(p, n) bitflush(p, n)
#define align_get_bits(p) bitbytealign(p)

#define bitload(p)								\
	if((p)->bitpos >= 8) {						\
		int bitpos = (p)->bitpos;				\
		const uint8_t* bitptr = (p)->bitptr;	\
		int bits = (p)->bits;					\
		do {									\
			bits = (bits << 8) | *bitptr++;		\
			bitpos -= 8;						\
		} while (bitpos >= 8);					\
		(p)->bits = bits;						\
		(p)->bitptr = bitptr;					\
		(p)->bitpos = bitpos;					\
	}
/*
STIN void bitload(bitstream* p)
{
	int bitpos = p->bitpos;

	if (bitpos >= 8) {
		const uint8_t* bitptr = p->bitptr;
		int bits = p->bits;
		do {
			bits = (bits << 8) | *bitptr++;
			bitpos -= 8;
		} while (bitpos >= 8);
		p->bits = bits;
		p->bitptr = bitptr;
	}											
	p->bitpos = bitpos;
}
*/
/*
STIN unsigned int get_bits(bitstream *p, int n)
{
	bitload(p);
	return bitget(p, n);
}
*/
static void init_get_bits(bitstream* p,const uint8_t *stream, int len)
{
	p->bits = 0;
	p->bitpos = 32;
	p->bitptr = stream;
	p->bitend = stream+(len >> 3)+4;
	bitload(p);
}

static inline int get_vlc(GetBitContext *s, VLC *vlc)
{
    int code;
    VLC_TYPE (*table)[2]= vlc->table;
    int n, index, nb_bits;
    int bits = vlc->bits;

    bitload(s);
    index= bitshow(s, bits);
    code = table[index][0];
    n    = table[index][1];

    if(n < 0){
		bitflush(s, bits);
		bitload(s);
        nb_bits = -n;

        index= bitshow(s, nb_bits) + code;
        code = table[index][0];
        n    = table[index][1];

        if(n < 0){
			bitflush(s, nb_bits);
			bitload(s);
            nb_bits = -n;

            index= bitshow(s, nb_bits) + code;
            code = table[index][0];
            n    = table[index][1];
        }
    }
    bitflush(s, n);

    return code;
}
#else

#define bitload(p)

typedef struct GetBitContext {
    const uint8_t *buffer, *buffer_end;
    int index;
    int size_in_bits;
} GetBitContext;

/* used to avoid missaligned exceptions on some archs (alpha, ...) */
#ifdef ARCH_X86
#    define unaligned32(a) (*(uint32_t*)(a))
#else
#    ifdef __GNUC__
static inline uint32_t unaligned32(const void *v) {
    struct Unaligned {
	uint32_t i;
    } __attribute__((packed));

    return ((const struct Unaligned *) v)->i;
}
#    elif defined(__DECC)
static inline uint32_t unaligned32(const void *v) {
    return *(const __unaligned uint32_t *) v;
}
#    else
static inline uint32_t unaligned32(const void *v) {
    return *(const uint32_t *) v;
}
#    endif
#endif //!ARCH_X86

static inline int unaligned32_be(const void *v)
{
#ifdef CONFIG_ALIGN
	const uint8_t *p=v;
	return (((p[0]<<8) | p[1])<<16) | (p[2]<<8) | (p[3]);
#else
	return be2me_32( unaligned32(v)); //original
#endif
}

#   define MIN_CACHE_BITS 25

#   define OPEN_READER(name, gb)\
        int name##_index= (gb)->index;\
        int name##_cache= 0;\

#   define CLOSE_READER(name, gb)\
        (gb)->index= name##_index;\

#   define UPDATE_CACHE(name, gb)\
        name##_cache= unaligned32_be( ((uint8_t *)(gb)->buffer)+(name##_index>>3) ) << (name##_index&0x07);\

#   define SKIP_CACHE(name, gb, num)\
        name##_cache <<= (num);\

// FIXME name?
#   define SKIP_COUNTER(name, gb, num)\
        name##_index += (num);\

#   define SKIP_BITS(name, gb, num)\
        {\
            SKIP_CACHE(name, gb, num)\
            SKIP_COUNTER(name, gb, num)\
        }\

#   define LAST_SKIP_BITS(name, gb, num) SKIP_COUNTER(name, gb, num)
#   define LAST_SKIP_CACHE(name, gb, num) ;

#   define SHOW_UBITS(name, gb, num) NEG_USR32(name##_cache, num)

#   define GET_CACHE(name, gb) ((uint32_t)name##_cache)

static inline int get_bits_count(GetBitContext *s){
    return s->index;
}

// reads 0-17 bits.
// Note, the alt bitstream reader can read upto 25 bits, but the libmpeg2 reader cant
static inline unsigned int get_bits(GetBitContext *s, int n){
    register int tmp;
    OPEN_READER(re, s)
    UPDATE_CACHE(re, s)
    tmp= SHOW_UBITS(re, s, n);
    LAST_SKIP_BITS(re, s, n)
    CLOSE_READER(re, s)
    return tmp;
}

// shows 0-17 bits.
// Note, the alt bitstream reader can read upto 25 bits, but the libmpeg2 reader cant
static inline unsigned int show_bits(GetBitContext *s, int n){
    register int tmp;
    OPEN_READER(re, s)
    UPDATE_CACHE(re, s)
    tmp= SHOW_UBITS(re, s, n);
//    CLOSE_READER(re, s)
    return tmp;
}

static inline void skip_bits(GetBitContext *s, int n){
 //Note gcc seems to optimize this to s->index+=n for the ALT_READER :))
    OPEN_READER(re, s)
    UPDATE_CACHE(re, s)
    LAST_SKIP_BITS(re, s, n)
    CLOSE_READER(re, s)
}

void init_get_bits(GetBitContext *s,
                   const uint8_t *buffer, int buffer_size);

void align_get_bits(GetBitContext *s);

// if the vlc code is invalid and max_depth=1 than no bits will be removed
// if the vlc code is invalid and max_depth>1 than the number of bits removed
// is undefined
#define GET_VLC(code, name, gb, table, bits)\
{\
    int n, index, nb_bits;\
\
    index= SHOW_UBITS(name, gb, bits);\
    code = table[index][0];\
    n    = table[index][1];\
\
    if(n < 0){\
        LAST_SKIP_BITS(name, gb, bits)\
        UPDATE_CACHE(name, gb)\
\
        nb_bits = -n;\
\
        index= SHOW_UBITS(name, gb, nb_bits) + code;\
        code = table[index][0];\
        n    = table[index][1];\
        if(n < 0){\
            LAST_SKIP_BITS(name, gb, nb_bits)\
            UPDATE_CACHE(name, gb)\
\
            nb_bits = -n;\
\
            index= SHOW_UBITS(name, gb, nb_bits) + code;\
            code = table[index][0];\
            n    = table[index][1];\
        }\
    }\
    SKIP_BITS(name, gb, n)\
}

// deprecated, dont use get_vlc for new code, use get_vlc2 instead or use GET_VLC directly
static inline int get_vlc(GetBitContext *s, VLC *vlc)
{
    int code;
    VLC_TYPE (*table)[2]= vlc->table;

    OPEN_READER(re, s)
    UPDATE_CACHE(re, s)

    GET_VLC(code, re, s, table, vlc->bits)    

    CLOSE_READER(re, s)

    return code;
}

#endif // BITSTREAM

/* misc math functions */

extern const uint8_t ff_log2_tab[256];

static inline int av_log2(unsigned int v)
{
    int n;

    n = 0;
    if (v & 0xffff0000) {
        v >>= 16;
        n += 16;
    }
    if (v & 0xff00) {
        v >>= 8;
        n += 8;
    }
    n += ff_log2_tab[v];

    return n;
}

static inline int av_log2_16bit(unsigned int v)
{
    int n;

    n = 0;
    if (v & 0xff00) {
        v >>= 8;
        n += 8;
    }
    n += ff_log2_tab[v];

    return n;
}

#ifdef _LOW_ACCURACY_
#  define X(n) (((((n)>>22)+1)>>1) - ((((n)>>22)+1)>>9))
#  define LOOKUP_T const unsigned char
#else
#  define X(n) (n)
#  define LOOKUP_T const int32_t
#endif

extern const uint8_t ff_sqrt_tab[128];

static inline int ff_sqrt(int a)
{
    int ret=0;
    int s;
    int ret_sq=0;
    
    if(a<128) return ff_sqrt_tab[a];
    
    for(s=15; s>=0; s--){
        int b= ret_sq + (1<<(s*2)) + (ret<<s)*2;
        if(b<=a){
            ret_sq=b;
            ret+= 1<<s;
        }
    }
    return ret;
}
#endif /* COMMON_H */
