#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <stdarg.h>
#include <SDL.h>
#include <SDL_thread.h>
#include <unicodefont.h>
#ifdef GP2X
#include <linux/soundcard.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include "mmsp2_if.h"
#endif
#include "asfheader.h"
#include "common.h"
#include "soundtouch/SoundTouch.h"

/* GP2X button mapping */
enum MAP_KEY
{
	VK_UP         , // 0
	VK_UP_LEFT    , // 1
	VK_LEFT       , // 2
	VK_DOWN_LEFT  , // 3
	VK_DOWN       , // 4
	VK_DOWN_RIGHT , // 5
	VK_RIGHT      , // 6
	VK_UP_RIGHT   , // 7
	VK_START      , // 8
	VK_SELECT     , // 9
	VK_FL         , // 10
	VK_FR         , // 11
	VK_FA         , // 12
	VK_FB         , // 13
	VK_FX         , // 14
	VK_FY         , // 15
	VK_VOL_UP     , // 16
	VK_VOL_DOWN   , // 17
	VK_TAT          // 18
};

/* The screen surface, joystick device */
SDL_Surface *screen = NULL;
SDL_Joystick *joy = NULL;

soundtouch::SoundTouch SoundTouch;

extern "C" {
void *wma_decode_init(void);
int wma_decode_superframe(void *ctx,
                                 void *data, int *data_size,
                                 uint8_t *buf, int buf_size);
int wma_decode_end(void *ctx);

extern int use_noise_coding;
extern int use_exp_vlc;
};

int done = 0;
int decode_done = 0;
int video_on = 0;

void msg_printf(const char *text, ...);

uint64_t next_frame_count;
uint64_t next_keyframe_count;
int next_keyframe_index;
int count_per_sec;

SDL_Thread *asf_th = NULL;
SDL_mutex *asf_mutex = NULL;
SDL_cond *asf_cond = NULL;
SDL_cond *asf_audio_cond = NULL;
SDL_cond *asf_video_cond = NULL;
#define EnterRead() SDL_LockMutex(asf_mutex)
#define LeaveRead() SDL_UnlockMutex(asf_mutex)
#define WaitRead() SDL_CondWait(asf_cond, asf_mutex)
#define SignalRead() SDL_CondBroadcast(asf_cond)
#define WaitASFAudio() SDL_CondWait(asf_audio_cond, asf_mutex)
#define SignalASFAudio() SDL_CondSignal(asf_audio_cond)
#define WaitASFVideo() SDL_CondWait(asf_video_cond, asf_mutex)
#define SignalASFVideo() SDL_CondSignal(asf_video_cond)

int asf_thread(void *data)
{
	while(!done) {
		EnterRead();
		dprintf("af: enter\n");
		if(audbufsize) {
			//msg_printf("signal audio\n");
			dprintf("af: sigaud\n");
			SignalASFAudio();
			WaitRead();
			dprintf("af: retaud\n");
		} else if(vidbufsize) {
			//msg_printf("signal video\n");
			dprintf("af: sigvid\n");
			SignalASFVideo();
			WaitRead();
			dprintf("af: retvid\n");
		} else {
			//msg_printf("read file\n");
			dprintf("af: read\n");
			if(asfRead() == 0) done = 1;
			//if(vidbufsize) vidbufsize = 0;
		}
		dprintf("af: leave\n");
		LeaveRead();
	}
	msg_printf("asf_thread: Terminated.\n");

	return 0;
}

#define AUDIO_BUFSIZE 0x80000
#define AUDIO_MINIMUM 0x2000
char audiobuf[AUDIO_BUFSIZE];
int audiopos = 0;
int audiolen = 0;
int audiodone = 0;
int samples = 2048;
uint64_t updatecount = 0;
void *ctx;
SDL_Thread *audio_th = NULL;
SDL_mutex *audio_mutex = NULL;
SDL_cond *audio_cond = NULL;
#define EnterAudio() SDL_LockMutex(audio_mutex)
#define LeaveAudio() SDL_UnlockMutex(audio_mutex)
#define WaitAudio() SDL_CondWait(audio_cond, audio_mutex)
#define SignalAudio() SDL_CondSignal(audio_cond)

int videopos = 0;
int videolen = 0;
int videodecode = 0;
SDL_Thread *video_th = NULL;
SDL_Thread *display_th = NULL;
SDL_mutex *video_mutex = NULL;
SDL_cond *video_cond = NULL;
#define EnterVideo() SDL_LockMutex(video_mutex)
#define LeaveVideo() SDL_UnlockMutex(video_mutex)
#define WaitVideo() SDL_CondWait(video_cond, video_mutex)
#define SignalVideo() SDL_CondSignal(video_cond)

#define MAX_FRAME 0x1000
int vidhead, vidtail;
uint64_t vidcount[MAX_FRAME];
uint32_t vidpos[MAX_FRAME];
uint32_t vidsize[MAX_FRAME];
int vidkeyframe[MAX_FRAME];

int PrintStr(SDL_Surface *pScreen, int x, int y, const char *str);

#define UPDATE_SIZE 8192
char updatebuf[8192];
int updatesize = 0;
int sampleshift = 1;
int tempo = 100;
int last_tempo;
int needflush = 1;
int putcount = 0;
int putgap = 0;

void sound_update(void *data, unsigned char *ptr, int len)
{
	int size;
	int updates = 0;

	while(!decode_done && len) {
		EnterAudio();
		dprintf("su: enter\n");
		while(!decode_done && len) {
			if(updatesize) {
				size = updatesize;
				if(size > len) size = len;
				memcpy(ptr, updatebuf, size);
				ptr += size;
				len -= size;
				updatesize -= size;
				updates += size * tempo / 100;
				if(updatesize)
					memcpy(updatebuf, updatebuf + size, updatesize);
				continue;
			}

			if(audiopos >= (AUDIO_BUFSIZE)) {
				audiopos -= (AUDIO_BUFSIZE);
				if(audiopos) dprintf("audio queue error(pos): %d\n", audiopos);
			} else if((audiolen == audiopos) && needflush){
				if(done) {
					SoundTouch.flush();
					needflush = 0;
					//msg_printf("decode done...\n");
				} else {
					//msg_printf("Audio Buffer Empty...\n");
					break;
				}
			} else {
	            updatesize = SoundTouch.receiveSamples((soundtouch::SAMPLETYPE*)updatebuf, len >> sampleshift) << sampleshift;

	            if(!updatesize) {
					size = (audiolen >= audiopos) ? audiolen - audiopos : ((AUDIO_BUFSIZE) - audiopos);

					if(!size) {
						if(!needflush) {
							decode_done = 1;
							//msg_printf("sound update done...\n");
						}
					} else {
						if(!updatecount && !putgap) {
							putgap = putcount - (SoundTouch.numUnprocessedSamples() << sampleshift);
						} else {
							updates = putcount - (SoundTouch.numUnprocessedSamples() << sampleshift) - putgap - updatecount;
						}

						if(size > UPDATE_SIZE) size = UPDATE_SIZE;
						SoundTouch.putSamples((soundtouch::SAMPLETYPE*)(audiobuf + audiopos), size >> sampleshift);
						audiopos += size;
						putcount += size;
					}
				}
			}
		}
		dprintf("su: leave\n");
		LeaveAudio();
		dprintf("su: sig\n");
		SignalAudio();
	}

	//msg_printf("Samples / Updates: %d, %d\n", samples, updates);
	updatecount += updates;
	if(updatecount >= next_frame_count) SignalVideo();
}

int audio_thread(void *data)
{
	int size = 0, copy = 0, readbufsize = 0;
	uint8_t readbuf[AUDIO_BUFSIZE];
	char buf[AUDIO_BUFSIZE];
	char *p;
	audiolen = audiopos = readbufsize = 0;
	while(!done) {
		if(!readbufsize) {
			EnterRead();
			if(!audbufsize) {
				dprintf("au: wait\n");
				WaitASFAudio();
				dprintf("au: ret\n");
			}
			if(!done) {
				memcpy(readbuf, audbuf, audbufsize);
				readbufsize = audbufsize;
				audbufsize = 0;
			}
			LeaveRead();
			dprintf("au: sig\n");
			SignalRead();
		}

		if(readbufsize) {
			if(size == 0) {
				wma_decode_superframe(ctx, buf, &size, readbuf, readbufsize);
				//printf("wma decode: %d -> %d\n", readbufsize, size);
				p = buf;
				readbufsize = 0;
			}

			EnterAudio();
			dprintf("au: enter\n");
			copy = 0;
			while((copy <= 0) && size && !done) {
				copy = (AUDIO_BUFSIZE) - audiolen;
				if(audiopos == 0) copy--;
				if(audiopos > audiolen) copy = audiopos - audiolen - (audheader->nChannels << 1);
				if(copy > size) copy = size;
				if(copy <= 0) {
					dprintf("au: waitbuf\n");
					WaitAudio();
				}
			}
			if((copy > 0) && size && !done) {
				memcpy(audiobuf + audiolen, p, copy);
				audiolen += copy;
				p += copy;
				if(audiolen >= (AUDIO_BUFSIZE)) {
					audiolen -= AUDIO_BUFSIZE;
					if(audiolen) dprintf("audio queue error(len): %d\n", audiolen);
				}
				size -= copy;
			}
			dprintf("au: leave\n");
			LeaveAudio();
		}
	}
	msg_printf("audio_thread: Terminated.\n");

	return 0;
}

int display_thread(void *data)
{
	int decode_begin = 1;
	while(!decode_done) {
		EnterVideo();
		if((videodecode) && (updatecount >= next_frame_count)) {
#ifdef GP2X
	        run_FDC_and_display();
#endif
	        videodecode = 0;
		} else if(videodecode){
			WaitVideo();
		}

		while(!decode_done && !videodecode) {
			while(!decode_done && (vidtail == vidhead)) WaitVideo();
			if(decode_done) break;
			if(decode_begin && !vidkeyframe[vidtail]) {
				if(++vidtail == MAX_FRAME) vidtail = 0;
				continue;
			} else if(decode_begin) {
				decode_begin = 0;
				next_keyframe_index = -1;
			}

			if((next_keyframe_index > 0) && (updatecount >= next_keyframe_count)) {
				static int skip_count = 0;
				vidtail = next_keyframe_index;
				//msg_printf("Skip to next keyframe(%d).\n", skip_count++);
				next_keyframe_index = -1;
			}

#ifdef GP2X
			//printf("decode %d frame...\n", vidtail);
		    if (run_940_decoder(vidpos[vidtail], vidsize[vidtail]) < 0) {
		        msg_printf("run 940 error\n");
		        done = 1;
		        break;
		    }

		    //printf("read 940 statue...\n");
		    if (read_mp4d_940_status(&mp4d_disp) < 0)
		    {
		        msg_printf("MP_CMD_RUN failed => goto quit\n");
		        done = 1;
		        break;
		    }

		    if ((mp4d_disp.Command == MP_CMD_ACK) || (mp4d_disp.Command == MP_CMD_DISPLAY)) {
				next_frame_count = vidcount[vidtail];
				if(vidkeyframe[vidtail]) { // next_frame_count >= next_keyframe_count) {
					int i = vidtail + 1;
					if(i == MAX_FRAME) i = 0;
					while(!vidkeyframe[i] && (i != vidhead)) {
						if(++i == MAX_FRAME) i = 0;
					}
					if(i != vidhead) {
						next_keyframe_count = vidcount[i];
						next_keyframe_index = i;
					} else {
						next_keyframe_count = 0;
						next_keyframe_index = -1;
					}
				}
				videodecode = 1;
				//printf("decode complete(%d)...\n", vidtail);
		    } else if (mp4d_disp.Command == MP_CMD_FAIL) {
				printf("MP_CMD_FAIL ==> exit routine\n");
				done = 1;
				break;
		    }
#else
			next_frame_count = vidcount[vidtail];
			videodecode = 1;
#endif

			if(++vidtail == MAX_FRAME) vidtail = 0;
		}
        LeaveVideo();
	}
	msg_printf("display_thread: Terminated.\n");

	return 0;
}

int video_thread(void *data)
{
	int nextpos = 0;
	vidhead = vidtail = 0;
	vidsize[0] = 0;
	next_keyframe_index = -1;

	while(!done) {
		EnterRead();
		if(!vidbufsize) {
			dprintf("vi: wait\n");
			WaitASFVideo();
			dprintf("vi: ret\n");
		}
		if(!done) {
			if(nextpos + vidbufsize >= 0x1E0000) nextpos = 0;
#ifdef GP2X
			memcpy(dbuf + nextpos, vidbuf, vidbufsize);
#endif
			vidcount[vidhead] = (uint64_t)vidtime * count_per_sec / 1000;
			vidpos[vidhead] = nextpos;
			vidsize[vidhead] = vidbufsize;
			vidkeyframe[vidhead++] = iskeyframe;
			if(vidhead == MAX_FRAME) vidhead = 0;
			nextpos += vidbufsize;
#ifdef GP2X
			memset(dbuf + nextpos, 0, 2048);
#endif
			nextpos = (nextpos + 2048) & ~0x0f;
			vidbufsize = 0;
		}
		LeaveRead();
		dprintf("vi: sig\n");
		SignalRead();

		if(!videodecode)
			SignalVideo();
	}
	msg_printf("video_thread: Terminated.\n");

	return 0;
}

int last_volume = -1;
void setVolume(int volume)
{
#ifdef GP2X
	int mixer = open("/dev/mixer", O_RDWR);

	if(volume < 0) volume = 0;
	if(volume > 90) volume = 90;

	if(last_volume != volume) {
		last_volume = volume;

		if(mixer >= 0) {
			int vol = volume | (volume << 8);
			ioctl(mixer, SOUND_MIXER_WRITE_PCM, &vol);
			close(mixer);
			
			msg_printf("Set volume: %d\n", volume);
		}
	}
#endif
}

int DrawFont(SDL_Surface *pScreen, int x, int y, uint16_t code, uint16_t color)
{
	unsigned short fontdata[12];
	int height = 0, width = 0;
	uint16_t *ptr = (uint16_t *)pScreen->pixels;
	int pitch = pScreen->pitch >> 1;
	ptr += y * pitch + x; 

	GetFont(code, fontdata, &height, &width);
	if(width == 0) return 0;
	if((x + width) > 320) return -1;

	for(int cy=0; cy<height; cy++)
	{
		for(int cx=0; cx<width; cx++)
		{
			if(fontdata[cy] & (0x8000 >> cx))
			{
				ptr[cx] = color;
			}
		}
		ptr += pitch;
	}

	return width;
}

int PrintStr(SDL_Surface *pScreen, int x, int y, const char *str)
{
	const char *p = str;
	while(*p) {
		int w = DrawFont(pScreen, x, y, *p, 0xffff);
		x += w;
		if(x >= 320) break;
		p++;
	}
	return x;
}

int PrintWStr(SDL_Surface *pScreen, int x, int y, const uint16_t *str)
{
	const uint16_t *p = str;
	while(*p) {
		int w = DrawFont(pScreen, x, y, *p, 0xffff);
		x += w;
		if(x >= 320) break;
		p++;
	}
	return x;
}

static const char *draw_text(const char *text, int &x, int &y, int color)
{
    const char *p = text;

    while(*p) {
		if((*p == 0x0D) || (*p == 0x0A)) {
			if((p[0] != p[1]) && ((p[1] == 0x0D) || (p[1] == 0x0A))) ++p;
			++p;
			x = 0;
			y += 12;
			break;
		} else if(*p == 0x20) {
            x += 6;
        } else if(*p >= 0x20) {
			int w = DrawFont(screen, x, y, *p, color);
			if(w == -1) {
				x = 0;
				y += 12;
				break;
			}
			x += w;
        }
        ++p;
    }

    return p;
}

void msg_printf(const char *text, ...)
{
	va_list arg;
	char buf[512];
    const char *p;
	static int x = 0, y = 0;

	va_start(arg, text);
	vsprintf(buf, text, arg);
	va_end(arg);
	
    printf("%s", buf);
    
    if(video_on) {
		if(x || y) { x = 0; y = 0; }
		return;
	}

    p = buf;
    do {
        if(y == 240) {
			SDL_Rect src = {0, 12, 320, 228};
			SDL_Rect dst = {0, 0, 320, 228};
	        SDL_Rect r = {0, 228, 320, 12};
			SDL_BlitSurface(screen, &src, screen, &dst);
	        SDL_FillRect(screen, &r, 0);
	        y -= 12;
        }
        p = draw_text(p, x, y, 0xFFFF);
    } while(*p);

    SDL_Flip(screen);
}

#define COLORKEY_RED 0x00
#define COLORKEY_GREEN 0x00
#define COLORKEY_BLUE 0x00

#ifdef GP2X
void RGBLayerInit()
{
	unsigned short *gp2xregs;
	int memfd = open("/dev/mem", O_RDWR | O_SYNC);
	if (memfd < 0) return;

	gp2xregs = (unsigned short *)mmap(0, 0x10000, PROT_READ | PROT_WRITE, MAP_SHARED, memfd, 0xC0000000);
    if (gp2xregs != NULL) {
		unsigned short mixmux = (gp2xregs[0x28DC >> 1] & 0xFFF0) | 1;

		gp2xregs[0x2902 >> 1] = (COLORKEY_GREEN << 8) | COLORKEY_BLUE;
		gp2xregs[0x2904 >> 1] = COLORKEY_RED;
		gp2xregs[0x28DC >> 1] = mixmux;
		gp2xregs[0x28DE >> 1] = 0x0F;

		munmap((void *)gp2xregs, 0x10000);
	}

	close(memfd);

	msg_printf("RGB Layer initialize...\n");
}

void RGBLayerUninit()
{
	unsigned short *gp2xregs;
	int memfd = open("/dev/mem", O_RDWR | O_SYNC);
	if (memfd < 0) return;

	gp2xregs = (unsigned short *)mmap(0, 0x10000, PROT_READ | PROT_WRITE, MAP_SHARED, memfd, 0xC0000000);
    if (gp2xregs != NULL) {
		unsigned short mixmux = (gp2xregs[0x28DC >> 1] & 0xFFF0);

		gp2xregs[0x28DC >> 1] = mixmux;
		gp2xregs[0x28DE >> 1] = 0x0F;

		munmap((void *)gp2xregs, 0x10000);
	}

	close(memfd);

	printf("RGB Layer uninitialize...\n");
}
#endif

void Terminate(void)
{
	if(ctx) wma_decode_end(ctx);
	ctx = NULL;

	asfClose();

	SDL_Quit();
#ifdef GP2X
	RGBLayerUninit();
	close_dualcpu_vpp_unmap();

	struct stat info;
	if( (lstat("./wmvplayer.gpe", &info) == 0) && S_ISREG(info.st_mode) ) {
		execl("./wmvplayer.gpe", "./wmvplayer.gpe", NULL);
	} else {
		chdir("/usr/gp2x");
		execl("/usr/gp2x/gp2xmenu", "/usr/gp2x/gp2xmenu", NULL);
	}
#endif
}

int main (int argc, char *argv[])
{
	/* Initialize SDL */
	if (SDL_Init (SDL_INIT_VIDEO | SDL_INIT_AUDIO | SDL_INIT_JOYSTICK) < 0) {
		fprintf (stderr, "Couldn't initialize SDL: %s\n", SDL_GetError ());
		exit (1);
	}
	atexit (Terminate);

	SDL_ShowCursor(SDL_DISABLE);

	/* Set 320x240 16-bits video mode */
	screen = SDL_SetVideoMode (320, 240, 16, SDL_SWSURFACE);
	if (screen == NULL) {
		fprintf (stderr, "Couldn't set 320x240x16 video mode: %s\n", SDL_GetError ());
		exit (2);
	}

	SDL_ShowCursor(SDL_DISABLE);

	/* Check and open joystick device */
	if (SDL_NumJoysticks() > 0) {
		joy = SDL_JoystickOpen(0);
		if(!joy) {
			fprintf (stderr, "Couldn't open joystick 0: %s\n", SDL_GetError ());
		}
	}

#ifdef GP2X
	/* Only use GP2X code here */
	if(argc != 2) {
		printf("File not found.\n");
		return 0;
	}
	asfOpen(argv[1]);
#endif

#ifdef WIN32
	/* Only use Windows code here */
	//asfOpen("wmv/Rayearth.asf");
	//asfOpen("wmv/mval.wmv");

	//asfOpen("conan08.wma");
	//asfOpen("beethoven.wma");
	asfOpen("wmv/twinbird.wmv");
	//asfOpen("C:\\Documents and Settings\\user\\My Documents\\Resource\\Sample\\WMV\\ef_demo.wmv");
	//asfOpen("ef_demo.wmv");
	//asfOpen("mok.wma");
	//asfOpen("192_44.wma");
	//asfOpen("256_44.wma");
	//asfOpen("320_44.wma");
	//asfOpen("vbr_44.wma");
	//asfOpen("vbr_48.wma");
	//asfOpen("ryukiop.wmv");

	//asfOpen("wma/sure promise.wma");
	//asfOpen("wmv/KiddyGradePromotion.asf");
	//asfOpen("wmv/silver.asf");
	//asfOpen("wmv/Dragonar.wmv");
	//asfOpen("wmv/vbr_44.wma");

	//asfOpen("wmv/0003 come on.wmv");
#endif

#ifdef GP2X
    if (mmsp2_940_interface_buffer_init() < 0) {
        printf("interface buffer init error\n");
        return -1;
    }

    if (open_dualcpu_vpp() < 0) {
        printf("dualcpu vpp device open error\n");
        return -1;
    }
#endif

	asf_mutex = SDL_CreateMutex();
	asf_cond = SDL_CreateCond();
	asf_audio_cond = SDL_CreateCond();
	asf_video_cond = SDL_CreateCond();
	audio_mutex = SDL_CreateMutex();
	audio_cond = SDL_CreateCond();
	video_mutex = SDL_CreateMutex();
	video_cond = SDL_CreateCond();

	int y = 0;
	int x;
	int video_ver = -1;

	if(vidheader) {
		msg_printf("Video Decoder Initialize...");
#ifdef GP2X
		video_ver = mmsp2_setup_940_interface_buffer(vidheader->biWidth, vidheader->biHeight, vidheader->biCompression);
	    if(video_ver < 0) {
#endif
			msg_printf("Failed\n");
#ifdef GP2X
	    } else {
		    int real_width  = ((vidheader->biWidth >> 4) << 4);
		    int real_height = ((vidheader->biHeight >> 4) << 4);
		    set_FDC_YUVB_plane(real_width, real_height);
			msg_printf("OK\n");
	
			msg_printf("Width: %d\n", vidheader->biWidth);
			msg_printf("Height: %d\n", vidheader->biHeight);
			msg_printf("WMV Codec version: %d\n", video_ver);
			RGBLayerInit();
		}
#endif
	}

	if(audheader) {	
		msg_printf("Audio Decoder Initialize...");
		ctx = wma_decode_init();
		if(ctx == NULL) {
			msg_printf("Failed\n");
			video_ver = -1;
		} else {
			msg_printf("OK\n");
		}
	}

	if(ctx && audheader) {
		SDL_AudioSpec spec;

		spec.format = AUDIO_S16;
		spec.channels = audheader->nChannels;
		spec.freq = audheader->nSamplesPerSec;
		spec.samples = 2048;
		if(spec.channels == 1) spec.samples >>= 1;
		if(spec.freq < 32000) spec.samples >>= 1;
		if(spec.freq < 16000) spec.samples >>= 1;
		if(video_ver >= 0)
			spec.samples >>= 1;

		spec.callback = sound_update;
		spec.userdata = NULL;

		samples = spec.samples;
		sampleshift = audheader->nChannels;
		count_per_sec = spec.freq << sampleshift;

		if (SDL_OpenAudio(&spec, NULL) < 0) {
			printf("Unable to open audio: %s.\n", SDL_GetError());
			return 0;
		}

		tempo = 100;
	    last_tempo = tempo;

	    SoundTouch.setSampleRate(audheader->nSamplesPerSec);
	    SoundTouch.setChannels(audheader->nChannels);
	    SoundTouch.setSetting(SETTING_USE_QUICKSEEK, 1);
	    SoundTouch.setSetting(SETTING_USE_AA_FILTER, 0);
	    SoundTouch.setTempoChange(tempo - 100);

		msg_printf("Samplerate: %d Hz\n", audheader->nSamplesPerSec);
		msg_printf("Channels: %s\n", audheader->nChannels == 1 ? "Mono" : "Stereo");
		msg_printf("Bitrate: %d Kbps\n", audheader->nAvgBytesPerSec * 8 / 1000);
		msg_printf("Sample size: %d\n", samples);
		msg_printf("Noise coding: %s\n", use_noise_coding ? "Yes" : "No");
		msg_printf("Exponent coding: %s\n", use_exp_vlc ? "VLC" : "LSP");
	}

/*
	if(title[0]) {
		x = PrintStr(screen, 0, y, "Title : ");
		PrintWStr(screen, x, y, title);
		y += 12;
	}

	if(author[0]) {
		x = PrintStr(screen, 0, y, "Author : ");
		PrintWStr(screen, x, y, author);
		y += 12;
	}
*/
	if(playtime) {
		int h, m, s;
		char str[32];
		h = playtime / 3600;
		m = (playtime / 60) % 60;
		s = playtime % 60;
		msg_printf("Time: %02d:%02d:%02d\n", h, m, s);
	}
	SDL_Flip(screen);

	if (video_ver < 0) {
		video_on = 0;
	} else {
		video_on = 1;
		msg_printf("");
		SDL_FillRect(screen, NULL, 0);
		SDL_Flip(screen);
	}
	
	if(ctx != NULL) {
		audio_th = SDL_CreateThread(audio_thread, NULL);
		if(video_on) {
			video_th = SDL_CreateThread(video_thread, NULL);
			display_th = SDL_CreateThread(display_thread, NULL);
		} else {
			asfVideoOff();
		}
		asf_th = SDL_CreateThread(asf_thread, NULL);
	}

	setVolume(50);
	done = (ctx == NULL);
	decode_done = done;
	if(done)
		msg_printf("Player initialize failed.\n");

	int nextcount = 0;
	int loop = 1;

	SDL_PauseAudio(0);

	while (loop && !decode_done)
	{
		SDL_Event event;

		/* Check for events */
		while (loop && !decode_done && SDL_PollEvent (&event))
		{
			switch (event.type)
			{
				case SDL_KEYDOWN:
					/* if press Ctrl + C, terminate program */
					switch(event.key.keysym.sym) {
						case SDLK_c:
							if(!(event.key.keysym.mod & (KMOD_LCTRL | KMOD_RCTRL))) break;
						case SDLK_ESCAPE:
							loop = 0;
							break;
						case ',':
							tempo -= 5;
						case '.':
							if(event.key.keysym.sym == '.')
								tempo += 5;
							if(tempo < 50) tempo = 50;
							if(tempo > 300) tempo = 300;
							if(last_tempo != tempo) {
								EnterAudio();
								msg_printf("Set tempo: %d %%\n", tempo);
								SoundTouch.setTempoChange(tempo - 100);
								LeaveAudio();
								last_tempo = tempo;
							}
							break;
					}
					break;
				case SDL_KEYUP:
					break;
				case SDL_JOYBUTTONDOWN:
					/* if press Start button, terminate program */
					switch(event.jbutton.button) {
						case VK_START:
							loop = 0;
							break;
						case VK_VOL_UP:
							setVolume(last_volume + 5);
							break;
						case VK_VOL_DOWN:
							setVolume(last_volume - 5);
							break;
						case VK_FL:
							tempo -= 5;
						case VK_FR:
							if(event.jbutton.button == VK_FR)
								tempo += 5;
							if(tempo < 50) tempo = 50;
							if(tempo > 300) tempo = 300;
							if(last_tempo != tempo) {
								EnterAudio();
								msg_printf("Set tempo: %d %%\n", tempo);
								SoundTouch.setTempoChange(tempo - 100);
								LeaveAudio();
								last_tempo = tempo;
							}
					}
					break;
				case SDL_JOYBUTTONUP:
					break;
				case SDL_QUIT:
					loop = 0;
					break;
				default:
					break;
			}
		}
		
		/* Processing */
		/*
		if(updatecount >= nextcount) {
			// Display time
			int h, m, s;
			char str[32];
			SDL_Rect rect = {0, y, 320, 12};

			int buf = (audiolen > audiopos) ? (audiolen - audiopos) : ((AUDIO_BUFSIZE) - audiopos + audiolen);
			//SDL_Rect bufrect = {0, y+12, 320, 12};
			//sprintf(str, "Audio Buffer: %d %%", buf * 100 / AUDIO_BUFSIZE);
			//SDL_FillRect(screen, &bufrect, 0);
			//PrintStr(screen, 0, y+12, str);

			s = updatecount / count_per_sec;
			h = s / 3600;
			m = (s / 60) % 60;
			s = s % 60;
			//sprintf(str, "Play: %02d:%02d:%02d\n", h, m, s);
			//SDL_FillRect(screen, &rect, 0);
			//PrintStr(screen, 0, y, str);
			//SDL_Flip(screen);
			//nextcount += count_per_sec;
			nextcount += (count_per_sec >> 2);
		}
		*/

		SDL_Delay(1);
	}

	//SDL_Delay(1000);
	EnterAudio();
	SDL_PauseAudio(1);
	video_on = 0;
	msg_printf("Exit.\n");
	done = 1;
	decode_done = 1;
	LeaveAudio();

	if(ctx != NULL) {
		SDL_CondBroadcast(asf_cond);
		SDL_CondBroadcast(audio_cond);
		SDL_CondBroadcast(asf_audio_cond);
		SDL_CondBroadcast(video_cond);
		SDL_CondBroadcast(asf_video_cond);

		SDL_WaitThread(asf_th, NULL);
		SDL_WaitThread(audio_th, NULL);
		if(video_ver >= 0) {
			SDL_WaitThread(video_th, NULL);
			SDL_WaitThread(display_th, NULL);
		}
		
		msg_printf("All thread shudown\n");
	}

	SDL_DestroyCond(asf_cond);
	SDL_DestroyCond(asf_audio_cond);
	SDL_DestroyCond(asf_video_cond);
	SDL_DestroyCond(audio_cond);
	SDL_DestroyCond(video_cond);
	SDL_DestroyMutex(asf_mutex);
	SDL_DestroyMutex(audio_mutex);
	SDL_DestroyMutex(video_mutex);

	return 0;
}
